import argparse
from julienne.filemodel import generate_files

# ===========================================================================

DESCRIPTION = """\
Generates multiple versions of a Python project, slicing each version based on
special tokens in the code comments. Lines or blocks of code can be in all
versions, or only show up in a subset.
"""

parser = argparse.ArgumentParser(description=DESCRIPTION)


HELP = """\
Configuration file that describes the Python project that is to be split into
versions.
"""

parser.add_argument('config_file', help=HELP)

parser.add_argument('-v', '--verbose', help="Print info while processing",
    action='store_true', default=False)

parser.add_argument('-i', '--info', help="Only show info, don't process",
    action='store_true', default=False)

parser.add_argument('-c', '--chapter', help="Only process a specific chapter",
    type=int, default=None)

# ===========================================================================

def main():
    args = parser.parse_args()
    generate_files(args.config_file, args.verbose, args.info, args.chapter)



import shutil
from difflib import Differ
from filecmp import cmp
from pathlib import Path
from unittest import TestCase

from julienne.filemodel import generate_files

# ============================================================================

class SampleFilesTestCase(TestCase):
    def test_e2e(self):
        here = Path(__file__).parent

        # Remove files from last run, create output directory
        output = here / Path('data/last_output')
        if output.exists():
            shutil.rmtree(output)

        output.mkdir(parents=True)

        # Generate results and compare to expected
        path = here / Path('data/sample.toml')
        generate_files(str(path))

        expected = here / Path('data/expected')

        # Validate directory listing is the same
        expected_set = {p.relative_to(expected) for p in expected.rglob('*')}
        output_set = {p.relative_to(output) for p in output.rglob('*')}
        self.assertEqual(expected_set, output_set)

        # Validate files themselves are the same
        for path in expected_set:
            expected_path = expected / path
            output_path = output / path

            if expected_path.is_dir():
                continue

            result = cmp(expected_path, output_path)
            if not result:
                d = Differ()
                expected_text = expected_path.read_text().splitlines(
                    keepends=True)
                output_text = output_path.read_text().splitlines(keepends=True)

                result = d.compare(expected_text, output_text)

                error = f"{path} did not match expectation.\n\n" + \
                    "".join(list(result))
                raise AssertionError(error)

    def test_failures(self):
        here = Path(__file__).parent
        path = here / Path('data/fail.toml')

        with self.assertRaises(Exception) as context:
            generate_files(str(path))

        error = context.exception
        self.assertIn("bad_code/bad_marker.py", str(error))
        self.assertIn("Unknown marker type", str(error))



from math import log, ceil
from pathlib import Path

import tomli

from julienne.parsers import range_token
from julienne.nodes import (DirNode, ConditionalDirNode, PoundFileNode,
    ConditionalPoundFileNode, ConditionalCopyOnlyFileNode, CopyOnlyFileNode,
    XMLFileNode, ConditionalXMLFileNode)

# ===========================================================================
# Utilities
# ===========================================================================

def _convert_path(base_path, path):
    if path.is_absolute():
        return path

    mixed = base_path / path
    return mixed.resolve()

def _traverse(chapter, node, cmd, *args):
    fn = getattr(node, cmd)
    fn(*args)

    for child in node.children:
        if isinstance(child, DirNode) and child.should_traverse(chapter):
            _traverse(chapter, child, cmd, *args)
        else:
            fn = getattr(child, cmd)
            fn(*args)

# ===========================================================================
# File Tree
# ===========================================================================

class FileTree:
    def __init__(self, config, base_path, base_dir, verbose=False):
        self.base_path = base_path
        self.base_dir = base_dir
        self.verbose = verbose

        # Find the Python style files that participate in the parsing
        self.pound_files = []
        globs = config.get('pound_globs', ['**/*.py', ])
        for pattern in globs:
            self.pound_files.extend(base_dir.glob(pattern))

        # Find the XML style files that participate in the parsing
        self.xml_files = []
        globs = config.get('xml_globs', ['**/*.xml', '**/*.htm', '**/*.html' ])
        for pattern in globs:
            self.xml_files.extend(base_dir.glob(pattern))

        # Find the files that specify a participation range
        self.ranged_files_map = {}
        for spec in config.get('ranged_files', {}).values():
            token = spec['range']
            for path in spec['files']:
                ranged_path = _convert_path(base_path, Path(path))
                self.ranged_files_map[ranged_path] = token

        # Find the directories to skip
        self.skip_dirs = []
        for dirname in config.get('skip_dirs', []):
            path = _convert_path(base_dir, Path(dirname))
            self.skip_dirs.append(path)

        # Other values from the config
        self.skip_patterns = config.get('skip_patterns', [])
        self.prefix = config.get('chapter_prefix', 'ch')
        self.chapter_map = config.get('chapter_map', {})

        # Build the file tree
        self.root = DirNode(self.base_dir)
        self._process_dir_node(self.root, base_dir)
        self._find_biggest()

        if self.verbose:
            print('\n** File tree:')
            _traverse(self.biggest, self.root, 'info')

    def _process_dir_node(self, parent, dir_path):
        for path in dir_path.iterdir():
            # Skip any paths that are in our ignore_substrings list
            skip = False
            for pattern in self.skip_patterns:
                if pattern in str(path):
                    skip = True
                    break

            if skip:
                # Path contained a skip pattern, don't process it
                if self.verbose:
                    print(f"Skipping {path} because of pattern={pattern}")
                continue

            try:
                if path.is_dir():
                    if path in self.skip_dirs:
                        if self.verbose:
                            print(f"Skipping {path} because it is in skip_dirs")

                        continue
                    elif path in self.ranged_files_map.keys():
                        token = self.ranged_files_map[path]
                        node = ConditionalDirNode(path, token)
                    else:
                        node = DirNode(path)

                    parent.children.append(node)
                    self._process_dir_node(node, node.path)
                else:
                    if path in self.pound_files:
                        if path in self.ranged_files_map.keys():
                            token = self.ranged_files_map[path]
                            node = ConditionalPoundFileNode(path, token)
                        else:
                            node = PoundFileNode(path)

                        node.parse_file()
                    elif path in self.xml_files:
                        if path in self.ranged_files_map.keys():
                            token = self.ranged_files_map[path]
                            node = ConditionalXMLFileNode(path, token)
                        else:
                            node = XMLFileNode(path)

                        node.parse_file()
                    elif path in self.ranged_files_map.keys():
                        token = self.ranged_files_map[path]
                        node = ConditionalCopyOnlyFileNode(path, token)
                    else:
                        node = CopyOnlyFileNode(path)

                    parent.children.append(node)
            except Exception as e:
                raise e.__class__(f"Error parsing {path}. " + str(e))

    def _find_biggest(self):
        # Need to find the biggest upper bound, might be in the nodes, in the
        # ranged map, or in the chapter map
        max_node = self._find_biggest_in_nodes(self.root, 1)
        max_chapter = int(sorted(self.chapter_map.keys())[0])

        max_ranged = 1
        for token in self.ranged_files_map.values():
            lower, upper = range_token(token)
            max_ranged = max(max_ranged, lower, upper)

        # Find biggest chapter and how many digits are in it for padding
        self.biggest = max(max_node, max_ranged, max_chapter)
        self.digits = int(ceil(log(self.biggest + 1, 10)))

    def _find_biggest_in_nodes(self, node, biggest=1):
        result = biggest
        if hasattr(node, "lower") and node.lower > result:
            result = node.lower
        if hasattr(node, "upper") and node.upper != -1 and node.upper > result:
            result = node.upper

        try:
            for child in node.children:
                if isinstance(child, DirNode):
                    subresult = self._find_biggest_in_nodes(child, result)
                    if subresult > result:
                        result = subresult
                elif isinstance(child, PoundFileNode):
                    if hasattr(child, 'highest') and child.highest > result:
                        result = child.highest
        except Exception as e:
            raise e.__class__(f"Problem occurred processing {child.path} " \
                + str(e))

        return result

    def generate(self, output_dir, single_chapter=None):
        parent_path = self.base_dir.parent

        if single_chapter is not None:
            output_path = output_dir / Path(f"ch{single_chapter}")
            _traverse(single_chapter, self.root, 'copy', single_chapter, 
                parent_path, output_path)
            return

        # Generate whole range of chapters
        for num in range(1, self.biggest + 1):
            print(f'Creating chapter {num}')

            # If this chapter is in the map, use the mapped suffix instead
            if str(num) in self.chapter_map:
                # Filename based on mapped suffix
                filename = f"{self.prefix}{self.chapter_map[str(num)]}"
            else:
                # Filename based chapter number, padded based on largest number
                filename = f"{self.prefix}{num:0{self.digits}}"

            # Call the "copy" command, traversing the tree to generate the
            # output
            output_path = output_dir / Path(filename)
            _traverse(num, self.root, 'copy', num, parent_path, output_path)

# ===========================================================================
# File Generation
# ===========================================================================

def generate_files(config_file, verbose=False, info_only=False, 
        single_chapter=None):
    path = Path(config_file)
    path.resolve()
    base_path = path.parent

    config = tomli.loads(path.read_text())

    # Check for / create output directory
    output_dir = _convert_path(base_path, Path(config['output_dir']))
    if output_dir.exists():
        if not output_dir.is_dir():
            raise AttributeError(('The value for "output_dir" in the config '
                'file pointed to an existing path that was not a directory'))
    else:
        output_dir.mkdir()

    # Check for source directory
    base_dir = Path(config['src_dir'])
    base_dir = _convert_path(base_path, Path(config['src_dir']))
    if not base_dir.is_dir():
        raise AttributeError(('The value for "src_dir" in the config file was '
            'not a valid directory'))

    if info_only:
        # If only showing info force verbose
        verbose = True

    # Build the tree and then generate the output
    tree = FileTree(config, base_path, base_dir, verbose)

    if info_only:
        # Don't generate chapters, you're done
        print('\n**Info only, no chapters generated')
        exit()

    if verbose:
        print('\n**Processing')
    tree.generate(output_dir, single_chapter)



import shutil

from julienne.parsers import (parse_pound_content, parse_xml_content, 
    range_token, chapter_in_range)

# ===========================================================================
# Directory Nodes
# ===========================================================================

class DirNode:
    """Node for directories in the file tree."""

    def __init__(self, path):
        self.path = path
        self.children = []

    def info(self):
        print('DirNode')
        print(f'   {self.path}')

    def should_traverse(self, chapter):
        return True

    def copy(self, chapter, base_path, output_path):
        rel = self.path.relative_to(base_path)
        new_dir = output_path / rel
        new_dir.mkdir(parents=True, exist_ok=True)


class ConditionalDirNode(DirNode):
    """Node for directories that participate conditionally."""

    def __init__(self, path, token):
        self.path = path
        self.children = []
        self.lower, self.upper = range_token(token)

    def info(self):
        print(f'ConditionalDirNode {self.lower} - {self.upper}')
        print(f'   {self.path}')

    def should_traverse(self, chapter):
        return chapter_in_range(chapter, True, self.lower, self.upper)

    def copy(self, chapter, base_path, output_path):
        rel = self.path.relative_to(base_path)
        new_dir = output_path / rel

        if chapter_in_range(chapter, True, self.lower, self.upper):
            new_dir.mkdir(parents=True, exist_ok=True)

# ===========================================================================
# Copy Only Nodes
# ===========================================================================

class CopyOnlyFileNode:
    """Node for files that get copied but not parsed"""

    def __init__(self, path):
        self.path = path

    def info(self):
        print('CopyOnlyFileNode')
        print(f'   {self.path}')

    def copy(self, chapter, base_path, output_path):
        rel = self.path.relative_to(base_path)
        dest = output_path / rel

        shutil.copy2(self.path, dest)


class ConditionalCopyOnlyFileNode:
    """Node for files that get copied conditionally, but not parsed."""


    def __init__(self, path, token):
        self.path = path
        self.lower, self.upper = range_token(token)

    def info(self):
        print(f'ConditionalCopyOnlyFileNode {self.lower} - {self.upper}')
        print(f'   {self.path}')

    def should_traverse(self, chapter):
        return chapter_in_range(chapter, True, self.lower, self.upper)

    def copy(self, chapter, base_path, output_path):
        rel = self.path.relative_to(base_path)
        dest = output_path / rel

        if chapter_in_range(chapter, True, self.lower, self.upper):
            shutil.copy2(self.path, dest)

# ===========================================================================
# Parsing Node Base Classes
# ===========================================================================

class _BaseFileNode:
    def __init__(self, path):
        self.path = path
        self._parser_fn = None

    def parse_file(self):
        ### Done as a separate step to make testing easier, allows for
        # testing the ._parse_content() method without having an actual file
        self._parse_content(self.path.read_text())

    def _parse_content(self, content):
        """Sets the list of parsed Line objects, one for each line in the 
        given string of content.

        :param content: string to parse
        """
        self.parser = self._parser_fn(content)
        self.lowest = self.parser.lowest
        self.highest = self.parser.highest
        self.all_conditional = self.parser.all_conditional

    def info(self):
        lowest = getattr(self, 'lowest', "Unset")
        highest = getattr(self, 'highest', "Unset")
        all_cond = getattr(self, 'all_conditional', "Unset")

        print(f'{self.__class__.__name__}', lowest, highest, all_cond)
        print(f'   {self.path}')

    def copy(self, chapter, base_path, output_path):
        rel = self.path.relative_to(base_path)

        # If the file is all conditional, only write it in the chapter range,
        # If the file is not all conditional, some parts will appear in every
        # chapter, so write it
        if not self.all_conditional or (self.all_conditional and \
                chapter_in_range(chapter, True, self.lowest, self.highest)):
            # Write file if within chapter range
            dest = output_path / rel
            with open(dest, "w") as f:
                for line in self.parser.lines:
                    content = line.get_content(chapter)
                    if content is not None:
                        f.write(content + "\n")

class ConditionalFileNodeMixin:
    def info(self):
        print(f'{self.__class__.__name__} {self.lower} - {self.upper}')
        print(f'   {self.path}')

    def copy(self, chapter, base_path, output_path):
        if not chapter_in_range(chapter, True, self.lower, self.upper):
            return

        super().copy(chapter, base_path, output_path)

# ===========================================================================
# Python File Nodes
# ===========================================================================

class PoundFileNode(_BaseFileNode):
    """Node for Python style files, those with comments that are a #. These
    are parsed and processed for every chapter."""
    def __init__(self, path):
        super().__init__(path)
        self._parser_fn = parse_pound_content


class ConditionalPoundFileNode(ConditionalFileNodeMixin, PoundFileNode):
    """Node for Python style files, those with comments that are a #. These
    are parsed and processed conditionally."""
    def __init__(self, path, token):
        self.lower, self.upper = range_token(token)
        super().__init__(path)

# ===========================================================================
# XML Nodes
# ===========================================================================

class XMLFileNode(_BaseFileNode):
    """Node for XML style files, those with comments that are a <!-- -->.
    These are parsed and processed for every chapter."""
    def __init__(self, path):
        super().__init__(path)
        self._parser_fn = parse_xml_content


class ConditionalXMLFileNode(ConditionalFileNodeMixin, XMLFileNode):
    """Node for XML style files, those with comments that are a <!-- -->.
    These are parsed and processed for conditionally."""
    def __init__(self, path, token):
        self.lower, self.upper = range_token(token)
        super().__init__(path)



from waelstow import noted_raise

from tests.base import BaseParserTestCase
from julienne.parsers import parse_pound_content

# ============================================================================

CODE_BLOCK1 = """\
#@+ 3-4
#@- e = "In chapters 3 to 4"  # inline comment
#@- f = "  as a block"\
"""

EXPECTED_BLOCK1 = [
    """e = "In chapters 3 to 4"  # inline comment""",
    """f = "  as a block"\
""",
]


CODE_BLOCK2 = """\
for x in range(10):
    #@+ 1-2 block header with comment
    #@- g= "In chapters 1 and 2"
    h = "In all chapters"\
"""

EXPECTED_BLOCK2 = [
    """for x in range(10):""",
    """    # block header with comment""",
    """    g= "In chapters 1 and 2"\
""",
    """    h = "In all chapters"\
"""
]

CODE_BLOCK3 = """\
#@[ 2- uncommented conditional block
def foo():
    print('Blah de blah')
#@]
"""

EXPECTED_BLOCK3 = [
    """# uncommented conditional block""",
    """def foo():""",
    """    print('Blah de blah')""",
]

# ----------------------------------------------------------------------------

class PoundParserTestCase(BaseParserTestCase):

    def test_line_parsing(self):
        # Test a comment
        text = "# This is a sample file"
        parser = parse_pound_content(text)
        self.assertParser(parser, False, [text,], 1, -1)

        # Test a regular line
        text = 'a = "In all chapters"   # inline comment'
        parser = parse_pound_content(text)
        self.assertParser(parser, False, [text,], 1, -1)

        # Test a full-range conditional line with inline comment
        text = 'b = "In chapters 1-3"   #@= 1-3 comment on conditional'
        expected = 'b = "In chapters 1-3"   # comment on conditional'
        parser = parse_pound_content(text)
        self.assertParser(parser, True, [expected,], 1, 3)

        # Test lower open range conditional line
        text = 'c = "In chapters 1-2"   #@= -2'
        expected = 'c = "In chapters 1-2"   '
        parser = parse_pound_content(text)
        self.assertParser(parser, True, [expected,], 1, 2)

        # Test upper open range conditional line
        text = 'd = "In chapters 2 on"  #@= 2-'
        expected = 'd = "In chapters 2 on"  '
        parser = parse_pound_content(text)
        self.assertParser(parser, True, [expected,], 2, -1)

    def test_bad_parsing(self):
        text = 'x = 3 #@'
        with self.assertRaises(ValueError):
            parse_pound_content(text)

        text = 'x = 3 #@*'
        with self.assertRaises(ValueError):
            parse_pound_content(text)

    def test_block_parsing(self):
        #--- Test a conditional block
        parser = parse_pound_content(CODE_BLOCK1)
        self.assertParser(parser, True, EXPECTED_BLOCK1, 3, 4)

        #--- Test a conditional block inside an indent
        parser = parse_pound_content(CODE_BLOCK2)
        self.assertParser(parser, False, EXPECTED_BLOCK2, 1, 2)

        # Validate the non-conditional parts
        with noted_raise("[chapter={chapter}]"):
            for chapter in range(1, 4):
                content = parser.lines[0].get_content(chapter)
                self.assertIsNotNone(content)
                content = parser.lines[3].get_content(chapter)
                self.assertIsNotNone(content)

                # Line[1] and Line[2] are conditional, showing in 
                # chapters 1-2 only
                if chapter <= 2:
                    content = parser.lines[1].get_content(chapter)
                    self.assertIsNotNone(content)
                    content = parser.lines[2].get_content(chapter)
                    self.assertIsNotNone(content)
                else:
                    content = parser.lines[1].get_content(chapter)
                    self.assertIsNone(content)
                    content = parser.lines[2].get_content(chapter)
                    self.assertIsNone(content)

        # --- Test an uncommented conditional block
        parser = parse_pound_content(CODE_BLOCK3)
        self.assertParser(parser, True, EXPECTED_BLOCK3, 2, -1)



from waelstow import noted_raise

from tests.base import BaseParserTestCase
from julienne.parsers import parse_xml_content

# ============================================================================

BLOCK1 = """\
<!--@+ 3-4
In chapters 3 to 4 <!-- inline comment -->
as a block
@+-->
"""

EXPECTED_BLOCK1 = [
    "In chapters 3 to 4 <!-- inline comment -->",
    """as a block\
""",
]

BLOCK2 = """\
Some unindented text
    <!--@+ 1-2 block header with comment
    In chapters 1 and 2
    @+-->
    some indented text\
"""

EXPECTED_BLOCK2 = [
    """Some unindented text""",
    """    <!-- block header with comment -->""",
    """    In chapters 1 and 2""",
    """    some indented text\
""",
]

BLOCK3 = """\
<!--@[ 2- uncommented conditional block -->
Some text
    some indented text
<!--@] -->
"""

EXPECTED_BLOCK3 = [
    """<!-- uncommented conditional block -->""", 
    """Some text""", 
    """    some indented text\
"""
]
# ----------------------------------------------------------------------------

class XMLParserTestCase(BaseParserTestCase):

    def test_line_parsing(self):
        # Test a comment
        text = "<!-- sample comment line -->"
        parser = parse_xml_content(text)
        self.assertParser(parser, False, [text,], 1, -1)

        # Test a regular line
        text = 'Sample text with <!-- inline comment -->'
        parser = parse_xml_content(text)
        self.assertParser(parser, False, [text,], 1, -1)

        # Test a full-range conditional line with inline comment
        text = 'In chapters 1-3   <!--@= 1-3 comment on conditional -->'
        expected = 'In chapters 1-3   <!-- comment on conditional -->'
        parser = parse_xml_content(text)
        self.assertParser(parser, True, [expected,], 1, 3)

        # Test lower open range conditional line
        text = 'In chapters 1-2   <!--@= -2 -->'
        expected = 'In chapters 1-2   '
        parser = parse_xml_content(text)
        self.assertParser(parser, True, [expected,], 1, 2)

        # Test upper open range conditional line
        text = 'In chapters 2 on  <!--@= 2- -->'
        expected = 'In chapters 2 on  '
        parser = parse_xml_content(text)
        self.assertParser(parser, True, [expected,], 2, -1)

    def test_bad_parsing(self):
        text = 'x = 3 <!--@ -->'
        with self.assertRaises(ValueError):
            parse_xml_content(text)

        text = 'x = 3 <!--@* -->'
        with self.assertRaises(ValueError):
            parse_xml_content(text)

    def test_block_parsing(self):
        #--- Test a conditional block
        parser = parse_xml_content(BLOCK1)
        self.assertParser(parser, True, EXPECTED_BLOCK1, 3, 4)

        #--- Test a conditional block inside an indent
        parser = parse_xml_content(BLOCK2)
        self.assertParser(parser, False, EXPECTED_BLOCK2, 1, 2)

        # Validate the non-conditional parts
        with noted_raise("[chapter={chapter}]"):
            for chapter in range(1, 4):
                content = parser.lines[0].get_content(chapter)
                self.assertIsNotNone(content)
                content = parser.lines[3].get_content(chapter)
                self.assertIsNotNone(content)

                # Line[1] and Line[2] are conditional, showing in 
                # chapters 1-2 only
                if chapter <= 2:
                    content = parser.lines[1].get_content(chapter)
                    self.assertIsNotNone(content)
                    content = parser.lines[2].get_content(chapter)
                    self.assertIsNotNone(content)
                else:
                    content = parser.lines[1].get_content(chapter)
                    self.assertIsNone(content)
                    content = parser.lines[2].get_content(chapter)
                    self.assertIsNone(content)

        # --- Test an uncommented conditional block
        parser = parse_xml_content(BLOCK3)
        self.assertParser(parser, True, EXPECTED_BLOCK3, 2, -1)
